// CPU GA Connect the Dots 

#include "stdio.h"

#define POP_SIZE 4096
#define NUM_POINTS 100
#define NUM_GENERATIONS 1000
#define MUTATE 0.05

struct Point {
    float   x;
    float   y;
    float distance(float x2, float y2)
    {
	 return sqrt((x-x2)*(x-x2) + (y-y2)*(y-y2));
    }
};

// Function prototypes
void savePoints(Point points[]);
void saveSolutions(int population[][NUM_POINTS], float fitness[]);
void swap(int *val1, int *val2);
void evalFitness(int i, int population[][NUM_POINTS],
		 Point points[], float fitness[]);
void crossover(int parent1,int parent2,int child, 
	       int population[][NUM_POINTS],
	       int tempPopulation[][NUM_POINTS]);

int main()
{
	Point points[NUM_POINTS];

	// ****** Generate random points
	srand(time(NULL));		// Seed random number generator
	for (int i=0; i<NUM_POINTS;i++)
	{
		points[i].x = ((float) rand() * 10) / RAND_MAX;
		points[i].y = ((float) rand() * 10) / RAND_MAX;
	}
	savePoints(points);		// Save points to a file


	// ****** Generate random initial population
	int population[POP_SIZE][NUM_POINTS];
	for (int i = 0; i < POP_SIZE; i++)
	{
		// Each individual index 0,1,2,3...NUM_POINTS-1
		for (int j = 0; j < NUM_POINTS; j++)
		{
			population[i][j]=j;
		}
		// Randomize by picking two and swapping them
		for (int j = 0; j < NUM_POINTS; j++)
		{
			int k = rand() % NUM_POINTS;
			swap(&population[i][k],&population[i][j]);
		}
	}

	// Main GA Loop
	float fitness[POP_SIZE];
	int tempPopulation[POP_SIZE][NUM_POINTS];

	for (int gen = 0; gen < NUM_GENERATIONS; gen++)
	{
		float totalFitness = 0;
		float minFitness;
		int best=0;
		// Fitness function for each individual
		for (int i = 0; i <POP_SIZE; i++)
		{
			evalFitness(i, population, points, fitness);
			totalFitness += fitness[i];
			if (i==0) 
				minFitness = fitness[i];
			else
			{
				if (fitness[i] < minFitness)
				{
					minFitness = fitness[i];
					best = i;
				}
			}
		}
		printf("Gen %d Best fitness = %f\n",gen, minFitness);

		// Elitism
		memcpy(&tempPopulation[0],&population[best],
			NUM_POINTS*sizeof(int));

		// Select two parents using simple tournament selection
		// until we fill up the next population
		for (int child = 1; child < POP_SIZE; child++)
		{

		  int parent1, parent2;
		  int r1, r2, r3;
		  float min;
		  r1= rand() % POP_SIZE;
		  r2= rand() % POP_SIZE;
		  r3= rand() % POP_SIZE;
		  min =  fitness[r1];
		  parent1 = r1;
		  if (min > fitness[r2])
		  {
		  	min = fitness[r2];
		  	parent1 = r2;
		  }
		  if (min > fitness[r3])
		  {
		  	min = fitness[r3];
		  	parent1 = r3;
		  }
		  r1= rand() % POP_SIZE;
		  r2= rand() % POP_SIZE;
		  r3= rand() % POP_SIZE;
		  min =  fitness[r1];
		  parent2 = r1;
		  if (min > fitness[r2])
		  {
		  	min = fitness[r2];
		  	parent2 = r2;
		  }
		  if (min > fitness[r3])
		  {
		  	min = fitness[r3];
		  	parent2 = r3;
		  }
		  crossover(parent1,parent2,child,population,tempPopulation);

		  // Mutate with a small chance
		  float r = (float) rand() / RAND_MAX;
		  if (r < MUTATE)
		  {
			int p1 = rand() % NUM_POINTS;
			int p2 = rand() % NUM_POINTS;
			swap(&tempPopulation[child][p1],
			     &tempPopulation[child][p2]);
 	 	  }
		}
		// Copy child pop back to the general population
		memcpy(population,tempPopulation,
			POP_SIZE*NUM_POINTS*sizeof(float));
	}
	saveSolutions(population,fitness);

	return 0;
}


void savePoints(Point points[])
{
 FILE *f;
 f = fopen("points.data","w");
 fprintf(f,"%d\n",NUM_POINTS);
 for (int i =0; i < NUM_POINTS; i++)
 {
	fprintf(f,"%f\t%f\n",points[i].x,points[i].y);
 }
 fprintf(f,"-999\n");
 fclose(f);
}

void saveSolutions(int population[][NUM_POINTS],float fitness[])
{
 FILE *f;
 f = fopen("solutions.data","w");
 fprintf(f,"%d\n",POP_SIZE);
 for (int i=0; i<POP_SIZE;i++)
 {
	for (int j=0; j<NUM_POINTS;j++)
		fprintf(f,"%d ",population[i][j]);
	fprintf(f," %f\n",fitness[i]);
 }
 fprintf(f,"-999\n");
 fclose(f);
}

void swap(int *val1, int *val2)
{
  int temp = *val1;
  *val1 = *val2;
  *val2 = temp;
}

void evalFitness(int i, int population[][NUM_POINTS],
		 Point points[], float fitness[])
{
  float total = 0;
  for (int p = 1; p < NUM_POINTS; p++)
  {
	Point p1 = points[population[i][p-1]];	
	Point p2 = points[population[i][p]];	
	total += p1.distance(p2.x,p2.y);
  }
  fitness[i] = total;
  return;   
}

// Copy random segment from one parent, then points in order
// from the second parent if not already in the child
void crossover(int parent1,int parent2,int child, 
	       int population[][NUM_POINTS],
	       int tempPopulation[][NUM_POINTS])
{
 char copied[NUM_POINTS];
 for (int i = 0; i < NUM_POINTS; i++)
	copied[i]=0;
 int start, end;
 start = rand() % NUM_POINTS;
 end = rand() % NUM_POINTS;
 if (start > end)
	swap(&start,&end);
 // Copy from start..end from parent1
 for (int i = start; i <= end; i++)
 {
	tempPopulation[child][i]=population[parent1][i];
	copied[population[parent1][i]]=1;
 }
 // Copy everything else from parent2
 int j = 0;
 for (int i = 0; i < NUM_POINTS; i++)
 {
 	if (j==start) j=end+1;
	if (j == NUM_POINTS)
	{
		break;
	}
	if (!copied[population[parent2][i]])
	{
		tempPopulation[child][j] = population[parent2][i];
		j++;
	}
 }
}

